import os

from django.core.wsgi import get_wsgi_application

os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'demoproj.settings')

# This application object is used by any WSGI server configured to use this
# file. This includes Django's development server, if the WSGI_APPLICATION
# setting points here.

application = get_wsgi_application()



import logging
import threading
import uuid
from typing import Callable, Union

from django.apps import apps
from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest, HttpResponse

from django_guid.config import settings

logger = logging.getLogger('django_guid')


class GuidMiddleware(object):
    """
    Gets a GUID from a request header, or generates a GUID if none is found, for each incoming request.
    Stored GUIDs are accessible from anywhere in the Django app.
    """

    _guid = {}

    def __init__(self, get_response: Callable) -> None:
        """
        One-time configuration and initialization.
        """
        self.get_response = get_response

        # This logic cannot be moved to config.py because apps are not yet initialized when that is executed
        if not apps.is_installed('django_guid'):
            raise ImproperlyConfigured('django_guid must be in installed apps')

    def __call__(self, request: HttpRequest) -> Union[HttpRequest, HttpResponse]:
        """
        Fetches the current thread ID from the pool and stores the GUID in the _guid class variable,
        with the thread ID as the key.

        Deletes the GUID from the object unless settings are overwritten.

        :param request: HttpRequest from Django
        :return: Passes on the request or response to the next middleware
        """
        # Process request and store the GUID on the thread
        self.set_guid(self._get_id_from_header(request))

        # Run all integrations
        for integration in settings.INTEGRATIONS:
            logger.debug('Running integration: `%s`', integration.identifier)
            integration.run(guid=self.get_guid())

        # ^ Code above this line is executed before the view and later middleware
        response = self.get_response(request)

        if settings.RETURN_HEADER:
            response[settings.GUID_HEADER_NAME] = self.get_guid()  # Adds the GUID to the response header
            if settings.EXPOSE_HEADER:
                response['Access-Control-Expose-Headers'] = settings.GUID_HEADER_NAME

        # Run tear down for all the integrations
        for integration in settings.INTEGRATIONS:
            logger.debug('Running tear down for integration: `%s`', integration.identifier)
            integration.cleanup()

        return response

    @classmethod
    def get_guid(cls, default: str = None) -> str:
        """
        Fetches the GUID of the current thread from _guid.
        If no value has been set for the current thread yet, a default value is returned.

        :param default: Optional value to return if no GUID has been set on the current thread.
        :return: GUID or default.
        """
        return cls._guid.get(threading.current_thread(), default)

    @classmethod
    def set_guid(cls, guid: str) -> None:
        """
        Assigns a GUID to the thread.

        :param guid: The GUID being assigned
        :return: None
        """
        cls._guid[threading.current_thread()] = guid

    @classmethod
    def delete_guid(cls) -> None:
        """
        Delete the thread's GUID.

        :return: None
        """
        guid = cls.get_guid()
        if guid:
            logger.debug('Deleting %s from _guid', guid)
            cls._guid.pop(threading.current_thread(), None)

    @staticmethod
    def _generate_guid() -> str:
        """
        Generates an UUIDv4/GUID as a string.

        :return: GUID
        """
        return uuid.uuid4().hex

    @staticmethod
    def _validate_guid(original_guid: str) -> bool:
        """
        Validates a GUID.

        :param original_guid: string to validate
        :return: bool
        """
        try:
            return bool(uuid.UUID(original_guid, version=4).hex)
        except ValueError:
            logger.warning('Failed to validate GUID %s', original_guid)
            return False

    def _get_correlation_id_from_header(self, request: HttpRequest) -> str:
        """
        Returns either the provided GUID or a new one depending on if the provided GUID is valid or not.

        :param request: HttpRequest object
        :return: GUID
        """
        given_guid = str(request.headers.get(settings.GUID_HEADER_NAME))
        if not settings.VALIDATE_GUID:
            logger.debug('Returning ID from header without validating it as a GUID')
            return given_guid
        elif settings.VALIDATE_GUID and self._validate_guid(given_guid):
            logger.debug('%s is a valid GUID', given_guid)
            return given_guid
        else:
            new_guid = self._generate_guid()
            logger.info('%s is not a valid GUID. New GUID is %s', given_guid, new_guid)
            return new_guid

    def _get_id_from_header(self, request: HttpRequest) -> str:
        """
        Checks if the request contains the header specified in the Django settings.
        If it does, we fetch the header and attempt to validate the contents as GUID.
        If no header is found, we generate a GUID to be injected instead.

        :param request: HttpRequest object
        :return: GUID
        """
        header = request.headers.get(settings.GUID_HEADER_NAME)  # Case insensitive headers.get added in Django2.2
        if header:
            logger.info('%s found in the header: %s', settings.GUID_HEADER_NAME, header)
            request.correlation_id = self._get_correlation_id_from_header(request)
        else:
            request.correlation_id = self._generate_guid()
            logger.info(
                'Header `%s` was not found in the incoming request. Generated new GUID: %s',
                settings.GUID_HEADER_NAME,
                request.correlation_id,
            )
        return request.correlation_id



from logging import Filter, LogRecord

from django_guid.middleware import GuidMiddleware


class CorrelationId(Filter):
    def filter(self, record: LogRecord) -> bool:
        """
        Determines that the specified record is to be logged.

        From the docs:
                Is the specified record to be logged? Returns 0 for no, nonzero for
                yes. If deemed appropriate, the record may be modified in-place.
        :param record: Log record
        :return: True
        """
        record.correlation_id = GuidMiddleware.get_guid()
        return True



import logging

logger = logging.getLogger(__name__)


def useless_function() -> bool:
    """
    Useless function to demonstrate a function log message.
    :return: True
    """
    logger.warning('Some warning in a function')
    return True



from django_guid.middleware import GuidMiddleware


def test_valid_guid():
    assert GuidMiddleware._validate_guid('07742cab407e4e8089ebfd191acbb752') is True


def test_is_valid_dashed_guid():
    assert GuidMiddleware._validate_guid('07742cab-407e-4e80-89eb-fd191acbb752') is True



"""
Django settings for demoproj project.

Generated by 'django-admin startproject' using Django 3.0.1.

For more information on this file, see
https://docs.djangoproject.com/en/3.0/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.0/ref/settings/
"""

import os

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

SECRET_KEY = 'secret'

DEBUG = True

ALLOWED_HOSTS = []

# Application definition

INSTALLED_APPS = [
    'django.contrib.admin',
    'django.contrib.auth',
    'django.contrib.contenttypes',
    'django.contrib.sessions',
    'django.contrib.messages',
    'django.contrib.staticfiles',
    'rest_framework',
    'django_guid',
]

MIDDLEWARE = [
    'django_guid.middleware.GuidMiddleware',  # <-- Add middleware here
    'django.middleware.security.SecurityMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = 'demoproj.urls'

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        # 'NAME': os.path.join(BASE_DIR, 'db.sqlite3'),
        'NAME': ':memory:',
    }
}

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

STATIC_URL = '/static/'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

# fmt: off

# Non required DJANGO_GUID settings (Set to default values here)
DJANGO_GUID = {
    'GUID_HEADER_NAME': 'Correlation-ID',
    'VALIDATE_GUID': True,
    'INTEGRATIONS': []
}

# Set up logging for the project
LOGGING = {
    'version': 1,
    'disable_existing_loggers': False,
    'filters': {
        'correlation_id': {
            '()': 'django_guid.log_filters.CorrelationId'  # <-- Add correlation ID to the filters
        }
    },
    'formatters': {
        'medium': {
            'format': '%(levelname)s %(asctime)s [%(correlation_id)s] %(name)s %(message)s'  # <-- Format the log string
        },
    },
    'handlers': {
        'console': {
            'class': 'logging.StreamHandler',
            'formatter': 'medium',
            'filters': ['correlation_id'],  # <-- Add filters to the handler
        },
    },
    'loggers': {
        'django': {
            'handlers': ['console'],  # <-- Specify handlers
            'level': 'INFO'
        },
        'demoproj': {
            'handlers': ['console'],  # <-- Specify handlers
            'level': 'DEBUG'
        },
        'django_guid': {
            'handlers': ['console'],
            'level': 'DEBUG'  # <-- Set to DEBUG to show log messages from DJANGO_GUID
        }
    }
}

# fmt: on



import pytest
from django.core.exceptions import ImproperlyConfigured


def test_missing_identifier(monkeypatch):
    """
    Tests that an exception is raised when identifier is missing.
    """
    from django_guid.integrations import SentryIntegration

    monkeypatch.setattr(SentryIntegration, 'identifier', None)
    with pytest.raises(ImproperlyConfigured, match='`identifier` cannot be None'):
        SentryIntegration()


def test_missing_run_method(monkeypatch, client):
    """
    Tests that an exception is raised when the run method has not been defined.
    """
    from django_guid.config import settings as guid_settings
    from django_guid.integrations import SentryIntegration

    monkeypatch.delattr(SentryIntegration, 'run')
    monkeypatch.setattr(guid_settings, 'INTEGRATIONS', [SentryIntegration()])
    with pytest.raises(ImproperlyConfigured, match='The integration `SentryIntegration` is missing a `run` method'):
        client.get('/api')


def test_run_method_not_accepting_kwargs(monkeypatch, client):
    """
    Tests that an exception is raised when the run method doesn't accept kwargs.
    """
    from django_guid.integrations import SentryIntegration
    from django.conf import settings
    from django_guid.config import Settings

    class BadIntegration(SentryIntegration):
        def run(self, guid):
            pass

    monkeypatch.setattr(settings, 'DJANGO_GUID', {'INTEGRATIONS': [BadIntegration()]})
    with pytest.raises(ImproperlyConfigured, match='Integration method `run` must accept keyword arguments '):
        Settings()


def test_cleanup_method_not_accepting_kwargs(monkeypatch, client):
    """
    Tests that an exception is raised when the run method doesn't accept kwargs.
    """
    from django_guid.integrations import SentryIntegration
    from django.conf import settings
    from django_guid.config import Settings

    class BadIntegration(SentryIntegration):
        def cleanup(self, guid):
            pass

    monkeypatch.setattr(settings, 'DJANGO_GUID', {'INTEGRATIONS': [BadIntegration()]})
    with pytest.raises(ImproperlyConfigured, match='Integration method `cleanup` must accept keyword arguments '):
        Settings()


def test_non_callable_methods(monkeypatch, subtests):
    """
    Tests that an exception is raised when any of the integration base methods are non-callable.
    """
    from django_guid.integrations import SentryIntegration
    from django.conf import settings
    from django_guid.config import Settings

    mock_integration = SentryIntegration()

    to_test = [
        {
            'function_name': 'cleanup',
            'error': 'Integration method `cleanup` needs to be made callable for `SentryIntegration`.',
        },
        {
            'function_name': 'run',
            'error': 'Integration method `run` needs to be made callable for `SentryIntegration`.',
        },
        {
            'function_name': 'setup',
            'error': 'Integration method `setup` needs to be made callable for `SentryIntegration`.',
        },
    ]

    for test in to_test:
        setattr(mock_integration, test.get('function_name'), 'test')
        monkeypatch.setattr(settings, 'DJANGO_GUID', {'INTEGRATIONS': [mock_integration]})
        with subtests.test(msg=f'Testing function {test.get("function_name")}'):
            with pytest.raises(ImproperlyConfigured, match=test.get('error')):
                Settings()


def test_base_class():
    """
    Test that a basic implementation of an integration works as expected.
    """
    from django_guid.integrations import Integration

    class MyCustomIntegration(Integration):
        identifier = 'My custom integration'

        def run(self, guid, **kwargs):
            pass

        def cleanup(self, **kwargs):
            pass

    stub_integration = MyCustomIntegration()
    assert stub_integration.setup() is None
    assert stub_integration.run('test') is None
    assert stub_integration.cleanup() is None



import logging
from typing import Optional

from django.core.signals import request_finished
from django.dispatch import receiver

from django_guid.middleware import GuidMiddleware

logger = logging.getLogger('django_guid')


@receiver(request_finished)
def delete_guid(sender: Optional[dict], **kwargs: dict) -> None:
    """
    Receiver function for when a request finishes.

    When a request is finished, delete a requests _guid reference to prevent memory leaks.

    :param sender: The sender of the signal. By documentation, we must allow this input parameter.
    :param kwargs: The request_finished signal does not actually send any kwargs, but Django will throw an error
        if we don't accept them. This is because at any point arguments could get added to the signal, and the receiver
        must be able to handle those new arguments.
    :return: None
    """
    logger.debug('Received signal `request_finished`')
    GuidMiddleware.delete_guid()



from django.core.exceptions import ImproperlyConfigured


class Integration(object):
    """
    Integration base class.
    """

    identifier = None  # The name of your integration

    def __init__(self) -> None:
        if self.identifier is None:
            raise ImproperlyConfigured('`identifier` cannot be None')

    def setup(self) -> None:
        """
        Holds validation and setup logic to be run when Django starts.
        """
        pass

    def run(self, guid: str, **kwargs) -> None:
        """
        Code here is executed in the middleware, before the view is called.
        """
        raise ImproperlyConfigured(f'The integration `{self.identifier}` is missing a `run` method')

    def cleanup(self, **kwargs) -> None:
        """
        Code here is executed in the middleware, after the view is called.
        """
        pass



import pytest
from django.core.exceptions import ImproperlyConfigured


@pytest.fixture
def mock_uuid(monkeypatch):
    class MockUUid:
        hex = '704ae5472cae4f8daa8f2cc5a5a8mock'

    monkeypatch.setattr('django_guid.middleware.uuid.uuid4', MockUUid)


def test_request_with_no_correlation_id(client, caplog, mock_uuid):
    """
    Tests a request without any correlation-ID in it logs the correct things.
    In this case, it means that the first log message should not have any correlation-ID in it, but the next two
    (from views and services.useless_file) should have.
    :param mock_uuid: Monkeypatch fixture for mocking UUID
    :param client: Django client
    :param caplog: caplog fixture
    """
    response = client.get('/')
    expected = [
        (
            'Header `Correlation-ID` was not found in the incoming request. Generated new GUID: 704ae5472cae4f8daa8f2cc5a5a8mock',
            None,
        ),
        ('This log message should have a GUID', '704ae5472cae4f8daa8f2cc5a5a8mock'),
        ('Some warning in a function', '704ae5472cae4f8daa8f2cc5a5a8mock'),
        ('Received signal `request_finished`', '704ae5472cae4f8daa8f2cc5a5a8mock'),
        ('Deleting 704ae5472cae4f8daa8f2cc5a5a8mock from _guid', '704ae5472cae4f8daa8f2cc5a5a8mock'),
    ]
    assert [(x.message, x.correlation_id) for x in caplog.records] == expected
    assert response['Correlation-ID'] == '704ae5472cae4f8daa8f2cc5a5a8mock'


def test_request_with_correlation_id(client, caplog):
    """
    Tests a request _with_ a correlation-ID in it logs the correct things.
    :param client: Django client
    :param caplog: caplog fixture
    """
    response = client.get('/', **{'HTTP_Correlation-ID': '97c304252fd14b25b72d6aee31565843'})
    expected = [
        ('Correlation-ID found in the header: 97c304252fd14b25b72d6aee31565843', None),
        ('97c304252fd14b25b72d6aee31565843 is a valid GUID', None),
        ('This log message should have a GUID', '97c304252fd14b25b72d6aee31565843'),
        ('Some warning in a function', '97c304252fd14b25b72d6aee31565843'),
        ('Received signal `request_finished`', '97c304252fd14b25b72d6aee31565843'),
        ('Deleting 97c304252fd14b25b72d6aee31565843 from _guid', '97c304252fd14b25b72d6aee31565843'),
    ]
    assert [(x.message, x.correlation_id) for x in caplog.records] == expected
    assert response['Correlation-ID'] == '97c304252fd14b25b72d6aee31565843'


def test_request_with_invalid_correlation_id(client, caplog, mock_uuid):
    """
    Tests that a request with an invalid GUID is replaced when VALIDATE_GUID is True.
    :param client: Django client
    :param caplog: Caplog fixture
    :param mock_uuid: Monkeypatch fixture for mocking UUID
    """
    response = client.get('/', **{'HTTP_Correlation-ID': 'bad-guid'})
    expected = [
        ('Correlation-ID found in the header: bad-guid', None),
        ('Failed to validate GUID bad-guid', None),
        ('bad-guid is not a valid GUID. New GUID is 704ae5472cae4f8daa8f2cc5a5a8mock', None),
        ('This log message should have a GUID', '704ae5472cae4f8daa8f2cc5a5a8mock'),
        ('Some warning in a function', '704ae5472cae4f8daa8f2cc5a5a8mock'),
        ('Received signal `request_finished`', '704ae5472cae4f8daa8f2cc5a5a8mock'),
        ('Deleting 704ae5472cae4f8daa8f2cc5a5a8mock from _guid', '704ae5472cae4f8daa8f2cc5a5a8mock'),
    ]
    assert [(x.message, x.correlation_id) for x in caplog.records] == expected
    assert response['Correlation-ID'] == '704ae5472cae4f8daa8f2cc5a5a8mock'


def test_request_with_invalid_correlation_id_without_validation(client, caplog, monkeypatch):
    """
    Tests that a request with an invalid GUID is replaced when VALIDATE_GUID is False.
    :param client: Django client
    :param caplog: Caplog fixture
    :param monkeypatch: Monkeypatch for django settings
    """
    from django_guid.config import settings as guid_settings

    monkeypatch.setattr(guid_settings, 'VALIDATE_GUID', False)
    client.get('/', **{'HTTP_Correlation-ID': 'bad-guid'})
    expected = [
        ('Correlation-ID found in the header: bad-guid', None),
        ('Returning ID from header without validating it as a GUID', None),
        ('This log message should have a GUID', 'bad-guid'),
        ('Some warning in a function', 'bad-guid'),
        ('Received signal `request_finished`', 'bad-guid'),
        ('Deleting bad-guid from _guid', 'bad-guid'),
    ]
    assert [(x.message, x.correlation_id) for x in caplog.records] == expected


def test_no_return_header_and_drf_url(client, caplog, monkeypatch, mock_uuid):
    """
    Tests that it does not return the GUID if RETURN_HEADER is false.
    This test also tests a DRF response, just to confirm everything works in both worlds.
    """
    from django_guid.config import settings as guid_settings

    monkeypatch.setattr(guid_settings, 'RETURN_HEADER', False)
    response = client.get('/api')
    expected = [
        (
            'Header `Correlation-ID` was not found in the incoming request. Generated new GUID: 704ae5472cae4f8daa8f2cc5a5a8mock',
            None,
        ),
        ('This is a DRF view log, and should have a GUID.', '704ae5472cae4f8daa8f2cc5a5a8mock'),
        ('Some warning in a function', '704ae5472cae4f8daa8f2cc5a5a8mock'),
        ('Received signal `request_finished`', '704ae5472cae4f8daa8f2cc5a5a8mock'),
        ('Deleting 704ae5472cae4f8daa8f2cc5a5a8mock from _guid', '704ae5472cae4f8daa8f2cc5a5a8mock'),
    ]
    assert [(x.message, x.correlation_id) for x in caplog.records] == expected
    assert not response.get('Correlation-ID')


def test_no_expose_header_return_header_true(client, monkeypatch, mock_uuid):
    """
    Tests that it does not return the Access-Control-Allow-Origin when EXPOSE_HEADER is set to False
    and RETURN_HEADER is True
    """
    from django_guid.config import settings as guid_settings

    monkeypatch.setattr(guid_settings, 'EXPOSE_HEADER', False)
    response = client.get('/api')
    assert not response.get('Access-Control-Expose-Headers')


def test_expose_header_return_header_true(client, monkeypatch, mock_uuid):
    """
    Tests that it does return the Access-Control-Allow-Origin when EXPOSE_HEADER is set to True
    and RETURN_HEADER is True
    """
    from django_guid.config import settings as guid_settings

    monkeypatch.setattr(guid_settings, 'EXPOSE_HEADER', True)
    response = client.get('/api')
    assert response.get('Access-Control-Expose-Headers')


def test_no_expose_header_return_header_false(client, monkeypatch, mock_uuid):
    """
    Tests that it does not return the Access-Control-Allow-Origin when EXPOSE_HEADER is set to False
    and RETURN_HEADER is False
    """
    from django_guid.config import settings as guid_settings

    monkeypatch.setattr(guid_settings, 'EXPOSE_HEADER', False)
    monkeypatch.setattr(guid_settings, 'RETURN_HEADER', False)
    response = client.get('/api')
    assert not response.get('Access-Control-Expose-Headers')


def test_expose_header_return_header_false(client, monkeypatch, mock_uuid):
    """
    Tests that it does not return the Access-Control-Allow-Origin when EXPOSE_HEADER is set to True
    and RETURN_HEADER is False
    """
    from django_guid.config import settings as guid_settings

    monkeypatch.setattr(guid_settings, 'EXPOSE_HEADER', True)
    monkeypatch.setattr(guid_settings, 'RETURN_HEADER', False)
    response = client.get('/api')
    assert not response.get('Access-Control-Expose-Headers')


def test_cleanup_signal(client, caplog, monkeypatch, mock_uuid):
    """
    Tests that a request cleans up a request after finishing.
    :param client: Django client
    :param caplog: Caplog fixture
    :param monkeypatch: Monkeypatch for django settings
    """
    from django_guid.config import settings as guid_settings

    monkeypatch.setattr(guid_settings, 'VALIDATE_GUID', False)
    client.get('/', **{'HTTP_Correlation-ID': 'bad-guid'})
    client.get('/', **{'HTTP_Correlation-ID': 'another-bad-guid'})

    expected = [
        # First request
        ('Correlation-ID found in the header: bad-guid', None),
        ('Returning ID from header without validating it as a GUID', None),
        ('This log message should have a GUID', 'bad-guid'),
        ('Some warning in a function', 'bad-guid'),
        ('Received signal `request_finished`', 'bad-guid'),
        ('Deleting bad-guid from _guid', 'bad-guid'),
        # Second request
        ('Correlation-ID found in the header: another-bad-guid', None),
        ('Returning ID from header without validating it as a GUID', None),
        ('This log message should have a GUID', 'another-bad-guid'),
        ('Some warning in a function', 'another-bad-guid'),
        ('Received signal `request_finished`', 'another-bad-guid'),
        ('Deleting another-bad-guid from _guid', 'another-bad-guid'),
    ]
    assert [(x.message, x.correlation_id) for x in caplog.records] == expected


def test_improperly_configured_if_not_in_installed_apps(client, monkeypatch):
    """
    Test that the app will fail if `is_installed('django_guid')` is `False`.
    """
    monkeypatch.setattr('django_guid.middleware.apps.is_installed', lambda x: False)
    with pytest.raises(ImproperlyConfigured, match='django_guid must be in installed apps'):
        client.get('/')



import logging

from django.core.exceptions import ImproperlyConfigured

from django_guid.integrations import Integration

logger = logging.getLogger('django_guid')


class SentryIntegration(Integration):
    """
    Ensures that each request's correlation ID is passed on to Sentry exception logs as a `transaction_id`.
    """

    identifier = 'SentryIntegration'

    def setup(self) -> None:
        """
        Verifies that the sentry_sdk dependency is installed.
        """
        # Makes sure the client has installed the `sentry_sdk` package, and that the header is appropriately named.
        try:
            import sentry_sdk  # noqa: F401
        except ModuleNotFoundError:
            raise ImproperlyConfigured(
                'The package `sentry-sdk` is required for extending your tracing IDs to Sentry. '
                'Please run `pip install sentry-sdk` if you wish to include this integration.'
            )

    def run(self, guid: str, **kwargs) -> None:
        """
        Sets the Sentry transaction_id.
        """
        import sentry_sdk

        with sentry_sdk.configure_scope() as scope:
            logger.debug('Setting Sentry transaction_id to %s', guid)
            scope.set_tag('transaction_id', guid)



import logging

from django.http import HttpRequest, JsonResponse
from rest_framework.decorators import api_view
from rest_framework.response import Response
from rest_framework.request import Request

from demoproj.services.useless_file import useless_function

logger = logging.getLogger(__name__)


def index_view(request: HttpRequest) -> JsonResponse:
    """
    Example view that logs a log and calls a function that logs a log.

    :param request: HttpRequest
    :return: JsonResponse
    """
    logger.info('This log message should have a GUID')
    useless_response = useless_function()
    return JsonResponse({'detail': f'It worked! Useless function response is {useless_response}'})


@api_view(('GET',))
def rest_view(request: Request) -> Response:
    """
    Example DRF view that logs a log and calls a function that logs a log.

    :param request: Request
    :return: Response
    """
    logger.info('This is a DRF view log, and should have a GUID.')
    useless_response = useless_function()
    return Response(data={'detail': f'It worked! Useless function response is {useless_response}'})



from django.apps import AppConfig


class DjangoGuidConfig(AppConfig):
    name = 'django_guid'

    def ready(self) -> None:
        """
        In order to avoid circular imports we import signals here.
        """
        from django_guid import signals  # noqa F401



from setuptools import find_packages, setup

from django_guid import __version__

with open('README.rst') as readme_file:
    readme = readme_file.read()
with open('CHANGELOG.rst') as changelog_file:
    changelog = changelog_file.read()

setup(
    name='django-guid',
    version=__version__,
    description='Middleware that makes a request GUID available from anywhere and injects it into your logs.',
    include_package_data=True,
    long_description=readme + '\n\n' + changelog,
    license='BSD',
    author='Jonas Krüger Svensson',
    author_email='jonas-ks@hotmail.com',
    url='https://github.com/JonasKs/django-guid',
    download_url='https://pypi.python.org/pypi/django-guid',
    packages=find_packages(exclude=['']),
    install_requires=['Django>=2.2'],
    keywords=['django', 'logging', 'request', 'web', 'uuid', 'guid', 'correlation', 'correlation-id'],
    platforms='OS Independent',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Internet :: WWW/HTTP :: WSGI',
        'Topic :: Software Development',
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



import pytest
from django.core.exceptions import ImproperlyConfigured


def test_sentry_integration(client, monkeypatch, caplog, mocker):
    """
    Tests the sentry integration
    """
    from sentry_sdk.scope import Scope
    from django_guid.integrations import SentryIntegration
    from django_guid.config import settings as guid_settings

    monkeypatch.setattr(guid_settings, 'INTEGRATIONS', [SentryIntegration()])
    mock_scope = mocker.patch.object(Scope, 'set_tag')

    client.get('/api', **{'HTTP_Correlation-ID': '97c304252fd14b25b72d6aee31565842'})
    expected = [
        (None, 'Correlation-ID found in the header: 97c304252fd14b25b72d6aee31565842'),
        (None, '97c304252fd14b25b72d6aee31565842 is a valid GUID'),
        ('97c304252fd14b25b72d6aee31565842', 'Running integration: `SentryIntegration`'),
        ('97c304252fd14b25b72d6aee31565842', 'Setting Sentry transaction_id to 97c304252fd14b25b72d6aee31565842'),
        ('97c304252fd14b25b72d6aee31565842', 'This is a DRF view log, and should have a GUID.'),
        ('97c304252fd14b25b72d6aee31565842', 'Some warning in a function'),
        ('97c304252fd14b25b72d6aee31565842', 'Running tear down for integration: `SentryIntegration`'),
        ('97c304252fd14b25b72d6aee31565842', 'Received signal `request_finished`'),
        ('97c304252fd14b25b72d6aee31565842', 'Deleting 97c304252fd14b25b72d6aee31565842 from _guid'),
    ]
    mock_scope.assert_called_with('transaction_id', '97c304252fd14b25b72d6aee31565842')
    assert [(x.correlation_id, x.message) for x in caplog.records] == expected


def test_sentry_validation(client, monkeypatch):
    """
    Tests that the package handles multiple header values by defaulting to one and logging a warning.
    """
    import sys
    from django_guid.integrations import SentryIntegration
    from django_guid.config import Settings
    from django.conf import settings

    # Mock away the sentry_sdk dependency
    sys.modules['sentry_sdk'] = None

    monkeypatch.setattr(settings, 'DJANGO_GUID', {'INTEGRATIONS': [SentryIntegration()]})
    with pytest.raises(
        ImproperlyConfigured,
        match='The package `sentry-sdk` is required for extending your tracing IDs to Sentry. '
        'Please run `pip install sentry-sdk` if you wish to include this integration.',
    ):
        Settings()



from warnings import warn

from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from django.utils.inspect import func_accepts_kwargs


class Settings(object):
    """
    Settings for django_guid read from the Django settings in `settings.py`.

    Inspired by django-auth-adfs from @jobec
    """

    def __init__(self) -> None:
        self.GUID_HEADER_NAME = 'Correlation-ID'
        self.VALIDATE_GUID = True
        self.RETURN_HEADER = True
        self.EXPOSE_HEADER = True
        self.INTEGRATIONS = []
        self.SKIP_CLEANUP = None  # Deprecated - to be removed in the next major version

        if hasattr(django_settings, 'DJANGO_GUID'):
            _settings = django_settings.DJANGO_GUID

            # Set user settings if provided
            for setting, value in _settings.items():
                if hasattr(self, setting):
                    setattr(self, setting, value)
                else:
                    raise ImproperlyConfigured(f'{setting} is not a valid setting for django_guid')

            if not isinstance(self.VALIDATE_GUID, bool):
                raise ImproperlyConfigured('VALIDATE_GUID must be a boolean')
            if not isinstance(self.GUID_HEADER_NAME, str):
                raise ImproperlyConfigured('GUID_HEADER_NAME must be a string')  # Note: Case insensitive
            if not isinstance(self.RETURN_HEADER, bool):
                raise ImproperlyConfigured('RETURN_HEADER must be a boolean')
            if not isinstance(self.EXPOSE_HEADER, bool):
                raise ImproperlyConfigured('EXPOSE_HEADER must be a boolean')
            if not isinstance(self.INTEGRATIONS, list):
                raise ImproperlyConfigured('INTEGRATIONS must be an array')

            for integration in self.INTEGRATIONS:

                # Make sure all integration methods are callable
                for method, name in [
                    (integration.setup, 'setup'),
                    (integration.run, 'run'),
                    (integration.cleanup, 'cleanup'),
                ]:
                    # Make sure the methods are callable
                    if not callable(method):
                        raise ImproperlyConfigured(
                            f'Integration method `{name}` needs to be made callable for `{integration.identifier}`.'
                        )

                    # Make sure the method takes kwargs
                    if name in ['run', 'cleanup'] and not func_accepts_kwargs(method):
                        raise ImproperlyConfigured(
                            f'Integration method `{name}` must '
                            f'accept keyword arguments (**kwargs) for `{integration.identifier}`.'
                        )

                # Run validate method
                integration.setup()

            if 'SKIP_CLEANUP' in _settings:
                warn(
                    'SKIP_CLEANUP was deprecated in v1.2.0, and no longer impacts package behaviour. '
                    'Please remove it from your DJANGO_GUID settings.',
                    DeprecationWarning,
                )


settings = Settings()



import pytest
from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured

from django_guid.config import Settings


def test_invalid_setting(monkeypatch):
    monkeypatch.setattr(django_settings, 'DJANGO_GUID', {'invalid_setting': 'some_value'})
    with pytest.raises(ImproperlyConfigured, match='invalid_setting is not a valid setting for django_guid'):
        Settings()


def test_invalid_guid(monkeypatch):
    monkeypatch.setattr(django_settings, 'DJANGO_GUID', {'VALIDATE_GUID': 'string'})
    with pytest.raises(ImproperlyConfigured, match='VALIDATE_GUID must be a boolean'):
        Settings()


def test_invalid_header_name(monkeypatch):
    monkeypatch.setattr(django_settings, 'DJANGO_GUID', {'GUID_HEADER_NAME': True})
    with pytest.raises(ImproperlyConfigured, match='GUID_HEADER_NAME must be a string'):
        Settings()


def test_invalid_skip_guid_setting(monkeypatch):
    """
    Assert that a deprecation warning is called when settings are instantiated with SKIP_CLEANUP == True or False
    """
    monkeypatch.setattr(django_settings, 'DJANGO_GUID', {'SKIP_CLEANUP': True})
    with pytest.deprecated_call():
        Settings()


def test_invalid_return_header_setting(monkeypatch):
    monkeypatch.setattr(django_settings, 'DJANGO_GUID', {'RETURN_HEADER': 'string'})
    with pytest.raises(ImproperlyConfigured, match='RETURN_HEADER must be a boolean'):
        Settings()


def test_invalid_expose_header_setting(monkeypatch):
    monkeypatch.setattr(django_settings, 'DJANGO_GUID', {'EXPOSE_HEADER': 'string'})
    with pytest.raises(ImproperlyConfigured, match='EXPOSE_HEADER must be a boolean'):
        Settings()


def test_valid_settings(monkeypatch):
    monkeypatch.setattr(
        django_settings,
        'DJANGO_GUID',
        {
            'VALIDATE_GUID': False,
            'GUID_HEADER_NAME': 'Correlation-ID-TEST',
            'RETURN_HEADER': False,
            'EXPOSE_HEADER': False,
        },
    )
    assert not Settings().VALIDATE_GUID
    assert Settings().GUID_HEADER_NAME == 'Correlation-ID-TEST'
    assert not Settings().RETURN_HEADER


def test_bad_integrations_type(monkeypatch):
    for item in [{}, '', 2, None, -2]:
        monkeypatch.setattr(django_settings, 'DJANGO_GUID', {'INTEGRATIONS': item})
        with pytest.raises(ImproperlyConfigured, match='INTEGRATIONS must be an array'):
            Settings()
